/* minimization of M/M#-convex function
 *  This is an implementation of steepest descent algorithm.
 *  The calculation cost is O(dimension^2) per step.
 *  cf. K. Murota: Discrete convex analysis (Japanese), 2001, p.227.
 *  cf. K. Murota: Discrete convex analysis, 2003, pp.281--283.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "odicon.h"

/* returns FALSE if no improvment was found. */
/* returns TRUE if improvment was found. */
/* '*plus' and '*minus' are the indexes of 'x[]'. */
static int best_improve(int dim, double func(int, int*), int x[],
                        int *plus,int *minus, int lo[],int up[], int natural) {
    int i, j, improved = FALSE;
    double f, min_f;

    iteration++;
    min_f = func(dim, x);
    for (i=0; i<dim+natural; i++) {
        x[i]--;
        if (i == dim || lo[i] <= x[i]) {
	    for (j=0; j<dim+natural; j++) {
	        if (j == i) continue;
		x[j]++;
		if (i == dim || x[i] <= up[i]) {
		    f = func(dim, x);  oracle_call++;
		    if (f < min_f) {
		        min_f = f;  *plus = j;  *minus = i;
			improved = TRUE;  improve_call++;
		    }
		}
		x[j]--;
	    }
	}
	x[i]++;
    }
    return improved;
}


/* returns the minimum value of func(x)    */
/* INPUT  func   : pointer to the function */
/* INPUT  init[] : the initial solution    */
/* OUTPUT init[] : the minimized solution  */
static double minimize(int dim, double func(int,int*), int init[],
                       int lower[], int upper[], int natural) {
    int *x, plus=0, minus=0, *lo, *up;

    counter_reset();
    if (!is_feasible(dim, func, init)) return PLUS_INFINITY;
    status_print(dim, func, init);

    assert(natural == 0 || natural == 1);
    x  = ivector_new(dim + natural);
    lo = ivector_new(dim + natural);
    up = ivector_new(dim + natural);
    ivector_copy(dim, x, init);
    ivector_copy(dim, lo, lower);
    ivector_copy(dim, up, upper);

    while (best_improve(dim, func, x, &plus, &minus, lo, up, natural)) {
        x[plus]++;
        x[minus]--;
        loop_print(dim, func, x);
    }
    ivector_copy(dim, init, x);
    free(x);

    counter_print();
    minimizer_print(dim, func, init);
    return func(dim, init);
}


double mconv_minimize(int dim, double func(int,int*),
		      int init[], int lower[], int upper[]) {
    return minimize(dim, func, init, lower, upper, 0);
}

double mgconv_minimize(int dim, double func(int,int*),
		       int init[], int lower[], int upper[]) {
    return minimize(dim, func, init, lower, upper, 1);
}
